@@ -89,6 +89,7 @@ int cf_listen_backlog;
 char *cf_unix_socket_dir;
 int cf_unix_socket_mode;
 char *cf_unix_socket_group;
+int cf_peer_id;
 
 int cf_pool_mode = POOL_SESSION;
 
@@ -114,6 +115,8 @@ char *cf_auth_file;
 char *cf_auth_hba_file;
 char *cf_auth_user;
 char *cf_auth_query;
+char *cf_auth_dbname;
+char *cf_track_extra_parameters;
 
 int cf_max_client_conn;
 int cf_default_pool_size;
@@ -148,6 +151,7 @@ usec_t cf_server_connect_timeout;
 usec_t cf_server_login_retry;
 usec_t cf_query_timeout;
 usec_t cf_query_wait_timeout;
+usec_t cf_cancel_wait_timeout;
 usec_t cf_client_idle_timeout;
 usec_t cf_client_login_timeout;
 usec_t cf_idle_transaction_timeout;
@@ -228,6 +232,7 @@ const struct CfLookup sslmode_map[] = {
 static const struct CfKey bouncer_params [] = {
 CF_ABS("admin_users", CF_STR, cf_admin_users, 0, ""),
 CF_ABS("application_name_add_host", CF_INT, cf_application_name_add_host, 0, "0"),
+CF_ABS("auth_dbname", CF_AUTHDB, cf_auth_dbname, 0, NULL),
 CF_ABS("auth_file", CF_STR, cf_auth_file, 0, NULL),
 CF_ABS("auth_hba_file", CF_STR, cf_auth_hba_file, 0, ""),
 CF_ABS("auth_query", CF_STR, cf_auth_query, 0, "SELECT usename, passwd FROM pg_shadow WHERE usename=$1"),
@@ -266,11 +271,13 @@ CF_ABS("max_db_connections", CF_INT, cf_max_db_connections, 0, "0"),
 CF_ABS("max_packet_size", CF_UINT, cf_max_packet_size, 0, "2147483647"),
 CF_ABS("max_user_connections", CF_INT, cf_max_user_connections, 0, "0"),
 CF_ABS("min_pool_size", CF_INT, cf_min_pool_size, 0, "0"),
+CF_ABS("peer_id", CF_INT, cf_peer_id, 0, "0"),
 CF_ABS("pidfile", CF_STR, cf_pidfile, CF_NO_RELOAD, ""),
 CF_ABS("pkt_buf", CF_INT, cf_sbuf_len, CF_NO_RELOAD, "4096"),
 CF_ABS("pool_mode", CF_LOOKUP(pool_mode_map), cf_pool_mode, 0, "session"),
 CF_ABS("query_timeout", CF_TIME_USEC, cf_query_timeout, 0, "0"),
 CF_ABS("query_wait_timeout", CF_TIME_USEC, cf_query_wait_timeout, 0, "120"),
+CF_ABS("cancel_wait_timeout", CF_TIME_USEC, cf_cancel_wait_timeout, 0, "10"),
 CF_ABS("reserve_pool_size", CF_INT, cf_res_pool_size, 0, "0"),
 CF_ABS("reserve_pool_timeout", CF_TIME_USEC, cf_res_pool_timeout, 0, "5"),
 CF_ABS("resolv_conf", CF_STR, cf_resolv_conf, CF_NO_RELOAD, ""),
@@ -290,7 +297,7 @@ CF_ABS("server_tls_cert_file", CF_STR, cf_server_tls_cert_file, 0, ""),
 CF_ABS("server_tls_ciphers", CF_STR, cf_server_tls_ciphers, 0, "fast"),
 CF_ABS("server_tls_key_file", CF_STR, cf_server_tls_key_file, 0, ""),
 CF_ABS("server_tls_protocols", CF_STR, cf_server_tls_protocols, 0, "secure"),
-CF_ABS("server_tls_sslmode", CF_LOOKUP(sslmode_map), cf_server_tls_sslmode, 0, "disable"),
+CF_ABS("server_tls_sslmode", CF_LOOKUP(sslmode_map), cf_server_tls_sslmode, 0, "prefer"),
 #ifdef WIN32
 CF_ABS("service_name", CF_STR, cf_jobname, CF_NO_RELOAD, NULL), /* alias for job_name */
 #endif
@@ -308,6 +315,7 @@ CF_ABS("tcp_keepidle", CF_INT, cf_tcp_keepidle, 0, "0"),
 CF_ABS("tcp_keepintvl", CF_INT, cf_tcp_keepintvl, 0, "0"),
 CF_ABS("tcp_socket_buffer", CF_INT, cf_tcp_socket_buffer, 0, "0"),
 CF_ABS("tcp_user_timeout", CF_INT, cf_tcp_user_timeout, 0, "0"),
+CF_ABS("track_extra_parameters", CF_STR, cf_track_extra_parameters, CF_NO_RELOAD, "IntervalStyle"),
 CF_ABS("unix_socket_dir", CF_STR, cf_unix_socket_dir, CF_NO_RELOAD, DEFAULT_UNIX_SOCKET_DIR),
 #ifndef WIN32
 CF_ABS("unix_socket_group", CF_STR, cf_unix_socket_group, CF_NO_RELOAD, ""),
@@ -331,6 +339,9 @@ static const struct CfSect config_sects [] = {
 	}, {
 		.sect_name = "users",
 		.set_key = parse_user,
+	}, {
+		.sect_name = "peers",
+		.set_key = parse_peer,
 	}, {
 		.sect_name = NULL,
 	}
@@ -385,6 +396,17 @@ static void set_dbs_dead(bool flag)
 	}
 }
 
+static void set_peers_dead(bool flag)
+{
+	struct List *item;
+	PgDatabase *db;
+
+	statlist_for_each(item, &peer_list) {
+		db = container_of(item, PgDatabase, head);
+		db->db_dead = flag;
+	}
+}
+
 /* Tells if the specified auth type requires data from the auth file. */
 static bool requires_auth_file(int auth_type)
 {
@@ -401,6 +423,7 @@ void load_config(void)
 	bool ok;
 
 	set_dbs_dead(true);
+	set_peers_dead(true);
 
 	/* actual loading */
 	ok = cf_load_file(&main_config, cf_config_file);
@@ -447,7 +470,7 @@ static void handle_sigterm(evutil_socket_t sock, short flags, void *arg)
 {
 	log_info("got SIGTERM, fast exit");
 	/* pidfile cleanup happens via atexit() */
-	exit(1);
+	exit(0);
 }
 
 static void handle_sigint(evutil_socket_t sock, short flags, void *arg)
@@ -561,6 +584,10 @@ static void go_daemon(void)
 {
 	int pid, fd;
 
+#ifdef WIN32
+	die("option --daemon (-d) is not supported on this platform");
+#endif
+
 	if (!cf_pidfile || !cf_pidfile[0])
 		die("daemon needs pidfile configured");
 
@@ -810,6 +837,16 @@ static void xfree(char **ptr_p)
 _UNUSED
 static void cleanup(void)
 {
+	/*
+	 * We don't want to cleanup when we're the target of a takeover, because
+	 * that would close the sockets that we hand over. There's no real clean
+	 * way to detect that we were the target, so the below check is rather
+	 * crude. But since this cleanup is only happening in builds with asserts
+	 * enabled anyway it seems fine.
+	 */
+	if (cf_pause_mode == P_SUSPEND && cf_shutdown == 2) {
+		return;
+	}
 	adns_free_context(adns);
 	adns = NULL;
 
@@ -831,6 +868,7 @@ static void cleanup(void)
 	xfree(&cf_unix_socket_dir);
 	xfree(&cf_unix_socket_group);
 	xfree(&cf_auth_file);
+	xfree(&cf_auth_dbname);
 	xfree(&cf_auth_hba_file);
 	xfree(&cf_auth_query);
 	xfree(&cf_auth_user);
@@ -857,6 +895,8 @@ static void cleanup(void)
 	xfree((char **)&cf_logfile);
 	xfree((char **)&cf_syslog_ident);
 	xfree((char **)&cf_syslog_facility);
+
+	xfree(&cf_track_extra_parameters);
 }
 
 /* boot everything */
@@ -937,6 +977,7 @@ int main(int argc, char *argv[])
 	init_objects();
 	load_config();
 	main_config.loaded = true;
+	init_var_lookup(cf_track_extra_parameters);
 	init_caches();
 	logging_prefix_cb = log_socket_prefix;
 
@@ -960,6 +1001,7 @@ int main(int argc, char *argv[])
 	admin_setup();
 
 	if (cf_reboot) {
+		log_warning("Online restart is deprecated, use so_reuseport instead");
 		if (check_old_process_unix()) {
 			takeover_part1();
 			did_takeover = true;
