@@ -31,6 +31,8 @@
 STATLIST(user_list);
 STATLIST(database_list);
 STATLIST(pool_list);
+STATLIST(peer_list);
+STATLIST(peer_pool_list);
 
 /* All locally defined users (in auth_file) are kept here. */
 struct AATree user_tree;
@@ -53,9 +55,12 @@ STATLIST(login_client_list);
 struct Slab *server_cache;
 struct Slab *client_cache;
 struct Slab *db_cache;
+struct Slab *peer_cache;
+struct Slab *peer_pool_cache;
 struct Slab *pool_cache;
 struct Slab *user_cache;
 struct Slab *iobuf_cache;
+struct Slab *var_list_cache;
 
 /*
  * libevent may still report events when event_del()
@@ -87,6 +92,7 @@ static void construct_client(void *obj)
 	memset(client, 0, sizeof(PgSocket));
 	list_init(&client->head);
 	sbuf_init(&client->sbuf, client_proto);
+	client->vars.var_list = slab_alloc(var_list_cache);
 	client->state = CL_FREE;
 }
 
@@ -97,6 +103,7 @@ static void construct_server(void *obj)
 	memset(server, 0, sizeof(PgSocket));
 	list_init(&server->head);
 	sbuf_init(&server->sbuf, server_proto);
+	server->vars.var_list = slab_alloc(var_list_cache);
 	server->state = SV_FREE;
 }
 
@@ -122,9 +129,11 @@ void init_objects(void)
 	aatree_init(&pam_user_tree, user_node_cmp, NULL);
 	user_cache = slab_create("user_cache", sizeof(PgUser), 0, NULL, USUAL_ALLOC);
 	db_cache = slab_create("db_cache", sizeof(PgDatabase), 0, NULL, USUAL_ALLOC);
+	peer_cache = slab_create("peer_cache", sizeof(PgDatabase), 0, NULL, USUAL_ALLOC);
+	peer_pool_cache = slab_create("peer_pool_cache", sizeof(PgPool), 0, NULL, USUAL_ALLOC);
 	pool_cache = slab_create("pool_cache", sizeof(PgPool), 0, NULL, USUAL_ALLOC);
 
-	if (!user_cache || !db_cache || !pool_cache)
+	if (!user_cache || !db_cache || !peer_cache || !peer_pool_cache || !pool_cache)
 		fatal("cannot create initial caches");
 }
 
@@ -140,6 +149,7 @@ void init_caches(void)
 	server_cache = slab_create("server_cache", sizeof(PgSocket), 0, construct_server, USUAL_ALLOC);
 	client_cache = slab_create("client_cache", sizeof(PgSocket), 0, construct_client, USUAL_ALLOC);
 	iobuf_cache = slab_create("iobuf_cache", IOBUF_SIZE, 0, do_iobuf_reset, USUAL_ALLOC);
+	var_list_cache = slab_create("var_list_cache", sizeof(struct PStr*) * get_num_var_cached(), 0, NULL, USUAL_ALLOC);
 }
 
 /* state change means moving between lists */
@@ -185,6 +195,7 @@ void change_client_state(PgSocket *client, SocketState newstate)
 	switch (client->state) {
 	case CL_FREE:
 		varcache_clean(&client->vars);
+		slab_free(var_list_cache, client->vars.var_list);
 		slab_free(client_cache, client);
 		break;
 	case CL_JUSTFREE:
@@ -255,6 +266,7 @@ void change_server_state(PgSocket *server, SocketState newstate)
 	switch (server->state) {
 	case SV_FREE:
 		varcache_clean(&server->vars);
+		slab_free(var_list_cache, server->vars.var_list);
 		slab_free(server_cache, server);
 		break;
 	case SV_JUSTFREE:
@@ -300,8 +312,25 @@ static int cmp_pool(struct List *i1, struct List *i2)
 	PgPool *p2 = container_of(i2, PgPool, head);
 	if (p1->db != p2->db)
 		return strcmp(p1->db->name, p2->db->name);
-	if (p1->user != p2->user)
+	if (p1->user != p2->user) {
+		if (p1->user == NULL) {
+			return 1;
+		}
+		if (p2->user == NULL) {
+			return -1;
+		}
 		return strcmp(p1->user->name, p2->user->name);
+	}
+	return 0;
+}
+
+/* compare pool names, for use with put_in_order */
+static int cmp_peer_pool(struct List *i1, struct List *i2)
+{
+	PgPool *p1 = container_of(i1, PgPool, head);
+	PgPool *p2 = container_of(i2, PgPool, head);
+	if (p1->db != p2->db)
+		return p1->db->peer_id - p2->db->peer_id;
 	return 0;
 }
 
@@ -313,6 +342,15 @@ static int cmp_user(struct List *i1, struct List *i2)
 	return strcmp(u1->name, u2->name);
 }
 
+/* compare db names, for use with put_in_order */
+static int cmp_peer(struct List *i1, struct List *i2)
+{
+	PgDatabase *db1 = container_of(i1, PgDatabase, head);
+	PgDatabase *db2 = container_of(i2, PgDatabase, head);
+	return db1->peer_id - db2->peer_id;
+}
+
+
 /* compare db names, for use with put_in_order */
 static int cmp_database(struct List *i1, struct List *i2)
 {
@@ -340,6 +378,25 @@ static void put_in_order(struct List *newitem, struct StatList *list,
 	statlist_append(list, newitem);
 }
 
+/* create new object if new, then return it */
+PgDatabase *add_peer(const char *name, int peer_id)
+{
+	PgDatabase *peer = find_peer(peer_id);
+
+	/* create new object if needed */
+	if (peer == NULL) {
+		peer = slab_alloc(peer_cache);
+		if (!peer)
+			return NULL;
+
+		list_init(&peer->head);
+		peer->peer_id = peer_id;
+		put_in_order(&peer->head, &peer_list, cmp_peer);
+	}
+
+	return peer;
+}
+
 /* create new object if new, then return it */
 PgDatabase *add_database(const char *name)
 {
@@ -474,6 +531,19 @@ PgUser *force_user(PgDatabase *db, const char *name, const char *passwd)
 	return user;
 }
 
+/* find an existing database */
+PgDatabase *find_peer(int peer_id)
+{
+	struct List *item;
+	PgDatabase *peer;
+	statlist_for_each(item, &peer_list) {
+		peer = container_of(item, PgDatabase, head);
+		if (peer->peer_id == peer_id)
+			return peer;
+	}
+	return NULL;
+}
+
 /* find an existing database */
 PgDatabase *find_database(const char *name)
 {
@@ -519,6 +589,7 @@ static PgPool *new_pool(PgDatabase *db, PgUser *user)
 
 	list_init(&pool->head);
 	list_init(&pool->map_head);
+	pool->orig_vars.var_list = slab_alloc(var_list_cache);
 
 	pool->user = user;
 	pool->db = db;
@@ -543,6 +614,38 @@ static PgPool *new_pool(PgDatabase *db, PgUser *user)
 	return pool;
 }
 
+
+/*
+ * create new peer pool object
+ *
+ * This pool should only be used to forward cancellations to other pgbouncers
+ * behind the same load balancer. The user field of this pool is NULL, because
+ * cancellations don't need a user.
+ */
+static PgPool *new_peer_pool(PgDatabase *db)
+{
+	PgPool *pool;
+
+	pool = slab_alloc(peer_pool_cache);
+	if (!pool)
+		return NULL;
+
+	list_init(&pool->head);
+	list_init(&pool->map_head);
+	pool->orig_vars.var_list = slab_alloc(var_list_cache);
+
+	pool->db = db;
+
+	statlist_init(&pool->new_server_list, "new_server_list");
+	statlist_init(&pool->waiting_cancel_req_list, "waiting_cancel_req_list");
+	statlist_init(&pool->active_cancel_req_list, "active_cancel_req_list");
+	statlist_init(&pool->active_cancel_server_list, "active_cancel_server_list");
+
+	/* keep pools in peer_id order to make stats faster */
+	put_in_order(&pool->head, &peer_pool_list, cmp_peer_pool);
+
+	return pool;
+}
 /* find pool object, create if needed */
 PgPool *get_pool(PgDatabase *db, PgUser *user)
 {
@@ -561,6 +664,17 @@ PgPool *get_pool(PgDatabase *db, PgUser *user)
 	return new_pool(db, user);
 }
 
+/* find pool object for the peer */
+PgPool *get_peer_pool(PgDatabase *db)
+{
+	if (!db)
+		return NULL;
+	if (!db->pool) {
+		db->pool = new_peer_pool(db);
+	}
+	return db->pool;
+}
+
 /* deactivate socket and put into wait queue */
 static void pause_client(PgSocket *client)
 {
@@ -906,6 +1020,7 @@ void disconnect_server(PgSocket *server, bool send_term, const char *reason, ...
 	case SV_TESTED:
 	case SV_USED:
 	case SV_IDLE:
+	case SV_BEING_CANCELED:
 		break;
 	case SV_LOGIN:
 		/*
@@ -955,13 +1070,41 @@ void disconnect_server(PgSocket *server, bool send_term, const char *reason, ...
 	free_scram_state(&server->scram_state);
 
 	server->pool->db->connection_count--;
-	server->pool->user->connection_count--;
+	if (server->pool->user) {
+		server->pool->user->connection_count--;
+	}
 
 	change_server_state(server, SV_JUSTFREE);
 	if (!sbuf_close(&server->sbuf))
 		log_noise("sbuf_close failed, retry later");
 }
 
+/*
+ * A wrapper around disconnect_client_sqlstate()
+ *
+ * The function disconnect_client_sqlstate() inherits the disconnect_client()
+ * content and add a new option that provides a specific SQLSTATE that is
+ * forwarded to client.  PgBouncer used to report SQLSTATE 08P01
+ * (protocol_violation) for all cases but it diverges from what Postgres
+ * reports in some cases.
+ */
+void disconnect_client(PgSocket *client, bool notify, const char *reason, ...)
+{
+	if (reason) {
+		char buf[128];
+		va_list ap;
+
+		va_start(ap, reason);
+		vsnprintf(buf, sizeof(buf), reason, ap);
+		va_end(ap);
+
+		disconnect_client_sqlstate(client, notify, NULL, buf);
+	} else {
+		disconnect_client_sqlstate(client, notify, NULL, reason);
+	}
+
+}
+
 /*
  * close client connection
  *
@@ -970,20 +1113,10 @@ void disconnect_server(PgSocket *server, bool send_term, const char *reason, ...
  * protocol and communication errors where sending a regular error
  * message is not possible.
  */
-void disconnect_client(PgSocket *client, bool notify, const char *reason, ...)
+void disconnect_client_sqlstate(PgSocket *client, bool notify, const char *sqlstate, const char *reason)
 {
 	usec_t now = get_cached_time();
 
-	if (reason) {
-		char buf[128];
-		va_list ap;
-
-		va_start(ap, reason);
-		vsnprintf(buf, sizeof(buf), reason, ap);
-		va_end(ap);
-		reason = buf;
-	}
-
 	if (cf_log_disconnections && reason)
 		slog_info(client, "closing because: %s (age=%" PRIu64 "s)", reason,
 			  (now - client->connect_time) / USEC);
@@ -1021,6 +1154,19 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason, ...)
 		break;
 	case CL_ACTIVE_CANCEL:
 	case CL_WAITING_CANCEL:
+		/*
+		 * During normal operation, cancel clients get closed because their
+		 * linked server finished sending the cancel request. But this is not
+		 * always the case. It's possible for the client to disconnect
+		 * itself. To avoid a reference to freed client object from the linked
+		 * server in such cases, we now unlink any still linked server.
+		 */
+		if (client->link) {
+			PgSocket *server = client->link;
+			server->link = NULL;
+			client->link = NULL;
+			disconnect_server(server, false, "client gave up on cancel request, so we also give up forwarding to server");
+		}
 		/*
 		 * If the cancel request is still linked to the server that it
 		 * cancelled (or wanted to cancel) a query from, this is the time to
@@ -1058,7 +1204,7 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason, ...)
 		 * don't send Ready pkt here, or client won't notice
 		 * closed connection
 		 */
-		send_pooler_error(client, false, true, reason);
+		send_pooler_error(client, false, sqlstate, true, reason);
 	}
 
 	free_scram_state(&client->scram_state);
@@ -1332,6 +1478,18 @@ void launch_new_connection(PgPool *pool, bool evict_if_needed)
 
 	max = pool_server_count(pool);
 
+	/*
+	 * Peer pools only have a single pool_size.
+	 */
+	if (pool->db->peer_id) {
+		if (max < pool_pool_size(pool))
+			goto force_new;
+
+		log_debug("launch_new_connection: peer pool full (%d >= %d)",
+				max, pool_pool_size(pool));
+		return;
+	}
+
 	/*
 	 * When a cancel request is queued allow connections up to twice the pool
 	 * size.
@@ -1411,7 +1569,9 @@ force_new:
 	pool->last_connect_time = get_cached_time();
 	change_server_state(server, SV_LOGIN);
 	pool->db->connection_count++;
-	pool->user->connection_count++;
+	if (pool->user) {
+		pool->user->connection_count++;
+	}
 
 	dns_connect(server);
 }
@@ -1496,6 +1656,57 @@ bool finish_client_login(PgSocket *client)
 	return true;
 }
 
+static void accept_cancel_request_for_peer(int peer_id, PgSocket *req) {
+	PgDatabase *peer = NULL;
+	PgPool *pool = NULL;
+	int ttl = req->cancel_key[7] & CANCELLATION_TTL_MASK;
+
+	if (ttl == 0) {
+		disconnect_client(req, false, "failed to forward cancel request because its TTL was exhausted");
+		return;
+	}
+
+	/*
+	 * Before forwarding the cancel key, we need to decrement the TTL. Now is
+	 * as a good a time as any to do so. We simply subtract 1 from the last
+	 * byte, since the TTL is stored in the least significant bits.
+	 */
+	req->cancel_key[7]--;
+
+	peer = find_peer(peer_id);
+	if (!peer) {
+		disconnect_client(req, false, "could not find peer to forward request to");
+		return;
+	}
+	log_debug("forwarding cancellation request to peer %d", peer_id);
+
+	/*
+	 * When using peering (multiple pgbouncers behind the same load
+	 * balancer), we may receive cancellation messages that were intended
+	 * for another peer via the load balancer. We propagate the
+	 * cancellation via the peer's pool, instead of the server pool.
+	 */
+	pool = get_peer_pool(peer);
+	if (!pool) {
+		disconnect_client(req, false, "out of memory");
+		return;
+	}
+
+	/*
+	 * Attach to the target pool and change state to waiting_cancel. This way
+	 * once a new connection is opened, it's used to forward the cancel
+	 * request.
+	 */
+	req->pool = pool;
+	change_client_state(req, CL_WAITING_CANCEL);
+
+	/*
+	 * Open a new connection over which the cancel request is forwarded to the
+	 * server.
+	 */
+	launch_new_connection(pool, /* evict_if_needed= */ true);
+}
+
 /*
  * Accepts a cancellation request, which will eventual cancel the query running
  * on the client that matches req->client_key
@@ -1505,9 +1716,34 @@ void accept_cancel_request(PgSocket *req)
 	struct List *pitem, *citem;
 	PgPool *pool = NULL;
 	PgSocket *server = NULL, *client, *main_client = NULL;
+	bool peering_enabled = false;
 
 	Assert(req->state == CL_LOGIN);
 
+	/*
+	 * PgBouncer peering
+	 */
+	peering_enabled = cf_peer_id > 0;
+	if (peering_enabled) {
+		/*
+		 * Extract the peer id from the cancel key
+		 */
+		int peer_id = req->cancel_key[0] + (req->cancel_key[1] << 8);
+		bool needs_forwarding_to_peer = cf_peer_id != peer_id;
+		if (needs_forwarding_to_peer) {
+			accept_cancel_request_for_peer(peer_id, req);
+			return;
+		}
+
+		/*
+		 * Set the last two bits of the cancel key to 1. This is necessary to
+		 * compare the key from the request to our stored cancel keys, because
+		 * the stored cancel keys always have these TTL bits set to 1.
+		 */
+		req->cancel_key[7] |= CANCELLATION_TTL_MASK;
+	}
+
+
 	/* find the client that has the same cancel_key as this request */
 	statlist_for_each(pitem, &pool_list) {
 		pool = container_of(pitem, PgPool, head);
@@ -1583,27 +1819,34 @@ bool forward_cancel_request(PgSocket *server)
 {
 	bool res;
 	PgSocket *req = first_socket(&server->pool->waiting_cancel_req_list);
+	bool forwarding_to_peer = server->pool->db->peer_id != 0;
 
 	Assert(req != NULL && req->state == CL_WAITING_CANCEL);
 	Assert(server->state == SV_LOGIN);
 
-	/*
-	 * In between accepting the cancel request and receiving an open connection
-	 * the query that was supposed to be cancelled has now completed. This
-	 * becomes a problem when the server is then reused for some other client.
-	 * Because this will mean that the cancel that is forwarded will cancel a
-	 * query from a completely different client than the client it was intended
-	 * for.
-	 */
-	if (!req->canceled_server) {
-		disconnect_client(req, false, "not sending cancel request for client that is now idle");
-		return false;
+	if (!forwarding_to_peer) {
+		/*
+		 * In between accepting the cancel request and receiving an open connection
+		 * the query that was supposed to be cancelled has now completed. This
+		 * becomes a problem when the server is then reused for some other client.
+		 * Because this will mean that the cancel that is forwarded will cancel a
+		 * query from a completely different client than the client it was intended
+		 * for.
+		 */
+		if (!req->canceled_server) {
+			disconnect_client(req, false, "not sending cancel request for client that is now idle");
+			return false;
+		}
 	}
 
 	server->link = req;
 	req->link = server;
 
-	SEND_CancelRequest(res, server, req->canceled_server->cancel_key);
+	if (forwarding_to_peer) {
+		SEND_CancelRequest(res, server, req->cancel_key);
+	} else {
+		SEND_CancelRequest(res, server, req->canceled_server->cancel_key);
+	}
 	if (!res) {
 		log_warning("sending cancel request failed: %s", strerror(errno));
 		disconnect_client(req, false, "failed to send cancel request");
@@ -1979,6 +2222,10 @@ void objects_cleanup(void)
 		db = container_of(item, PgDatabase, head);
 		kill_database(db);
 	}
+	statlist_for_each_safe(item, &peer_list, tmp) {
+		PgDatabase *peer = container_of(item, PgDatabase, head);
+		kill_peer(peer);
+	}
 
 	memset(&login_client_list, 0, sizeof login_client_list);
 	memset(&user_list, 0, sizeof user_list);
@@ -1993,10 +2240,16 @@ void objects_cleanup(void)
 	client_cache = NULL;
 	slab_destroy(db_cache);
 	db_cache = NULL;
+	slab_destroy(peer_cache);
+	peer_cache = NULL;
+	slab_destroy(peer_pool_cache);
+	peer_pool_cache = NULL;
 	slab_destroy(pool_cache);
 	pool_cache = NULL;
 	slab_destroy(user_cache);
 	user_cache = NULL;
 	slab_destroy(iobuf_cache);
 	iobuf_cache = NULL;
+	slab_destroy(var_list_cache);
+	var_list_cache = NULL;
 }
